#!/usr/bin/env python
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 3. Neither the name of the Author nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: build_unix_wrappers.py 169498 2009-08-31 15:13:31Z satskyse $
#
# Author: Sergey Satskiy
#

""" Build C++ Toolkit wrappers on UNIX """

import os, sys, os.path, re, tempfile
from optparse import OptionParser
from subprocess import Popen, PIPE
from traceback import format_tb


# disable pylint message about exception handling
# pylint: disable-msg=W0702
# pylint: disable-msg=W0704

def main():
    """ main entry point """

    parser = OptionParser( "%prog <version> <.../c++ directory> " \
                           "<library directory> <install directory>" )
    parser.add_option( "-v", "--verbose", action="store_true", default=False,
                       help="be verbose", dest="verbose" )
    options, args = parser.parse_args()

    if len( args ) != 4:
        return parserError( parser, "Incorrect number of arguments" )

    version, cppDir, libDir, installDir = args
    cppDir = os.path.abspath( cppDir )
    if not cppDir.endswith( "/" ):
        cppDir += "/"

    libDir = os.path.abspath( libDir )
    if not libDir.endswith( "/" ):
        libDir += "/"

    buildDir = libDir.replace( "/lib/", "/build/" )

    installDir = os.path.abspath( installDir )
    if not installDir.endswith( "/" ):
        installDir += "/"

    if options.verbose:
        print "Building C++ Toolkit wrappers on UNIX"
        print "Package version:   " + version
        print "Source directory:  " + cppDir
        print "Library directory: " + libDir
        print "Build directory:   " + buildDir
        print "Install directory: " + installDir

    if not os.path.exists( cppDir ):
        print >> sys.stderr, "Cannot find c++ directory. " \
                             "Expected it here: " + cppDir + ". Exiting"
        return 2

    if not os.path.exists( libDir ):
        print >> sys.stderr, "Cannot find library directory. " \
                             "Expected it here: " + libDir + ". Exiting"
        return 2

    if not os.path.exists( buildDir ):
        print >> sys.stderr, "Cannot find build directory. " \
                             "Expected it here: " + buildDir + ". Exiting"
        return 2

    if not os.path.exists( installDir ):
        print >> sys.stderr, "Cannot find install directory. " \
                             "Expected it here: " + installDir + ". Exiting"
        return 2


    # Build a list of all the toolkit header files
    headers = []
    toolkitHeaders( cppDir + "include", headers )


    # cd to the directory where the wrappers are going to be built
    wrappersTopDir = cppDir + "src/internal/cppcore/toolkit_wrappers/"
    if not os.path.exists( wrappersTopDir ):
        print >> sys.stderr, "Cannot find wrappers project directory. " \
                             "Expected it here: " + wrappersTopDir + ". Exiting"
        return 2
    currentDir = os.getcwd()
    os.chdir( wrappersTopDir )

    # Save the full list of the toolkit headers in a file
    try:
        f = open( "toolkit_headers", "w" )
        for item in headers:
            f.write( item + "\n" )
        f.close()
    except:
        pass

    # Exclude the files which should be ignored
    acceptedHeaders = excludeHeaders( headers, cppDir, options.verbose )

    # Save the full list of the toolkit headers in a file
    try:
        f = open( "accepted_toolkit_headers", "w" )
        for item in acceptedHeaders:
            f.write( item + "\n" )
        f.close()
    except:
        pass

    retCode = 0
    while True:

        # Take the SWIG files and check that everything is covered
        retCode = checkCoverage( acceptedHeaders )
        if retCode != 0:
            break

        # Generate version.i file
        retCode = generateVersionFile( version )
        if retCode != 0:
            break

        # Run SWIG
        retCode = runSWIG( cppDir + "include/", buildDir, options.verbose )
        if retCode != 0:
            break

        # Replace the 0-length .cpp file in the wrapper project dir and run make
        retCode = buildWrapperLibrary( buildDir, options.verbose )
        if retCode != 0:
            break

        # build ncbipy
        retCode = buildNCBIpy( buildDir, options.verbose )
        if retCode != 0:
            break

        # Copy the final libraries to the location where they are picked up
        # by the release framework to be sent back to the client
        retCode = copyReadyLibraries( libDir, installDir )
        if retCode != 0:
            break

        break

    os.chdir( currentDir )
    return retCode


def buildNCBIpy( buildDir, verbose ):
    """ builds ncbipy if python is selected """

    if not "python" in getTargetLanguages():
        return 0

    # Replace the source to the real one
    srcFileName = "python_wrappers/ncbipy.c.real"
    tgtFileName = "python_wrappers/ncbipy.c"

    if not os.path.exists( tgtFileName + ".bak" ):
        os.system( "cp " + tgtFileName + " " + tgtFileName + ".bak" )

    os.system( "cp -f " + srcFileName + " " + tgtFileName )

    # Run building the ncbipy
    cmdLine = "cd " + buildDir + "; make -f Makefile.flat ncbipy.exe"
    if verbose:
        print "ncbipy building command line: " + cmdLine

    retCode = os.system( cmdLine )
    if retCode != 0:
        print >> sys.stderr, "Error making ncbipy.\n Exiting."
        return 2

    return 0


def copyReadyLibraries( libDir, installDir ):
    """
    copies the ready libraries to the location where they are picked up
    and put into the final tar.gz
    """

    for lang in getTargetLanguages():
        srcFile = libDir + "lib" + lang + "_wrappers.so"
        if not os.path.exists( srcFile ):
            print >> sys.stderr, "Wrapper library is not found for " + lang + \
                                 ". Expected here: " + srcFile + "\nExiting."
            return 2

        targetDir = installDir + lang + "_wrappers/"
        if not os.path.exists( targetDir ):
            os.mkdir( targetDir )
        if not os.path.isdir( targetDir ):
            print >> sys.stderr, "Cannot create " + lang + \
                                 " wrappers directory.\nExiting."
            return 2
        if lang == "python":
            dstFile = targetDir + "_ncbi.so"
            os.system( "strip " + srcFile )
            os.system( "cp " + srcFile + " " + dstFile )
            os.system( "cp ncbi.py " + targetDir )

            ncbipyFile = libDir.replace( "/lib/", "/bin/" ) + "ncbipy"
            os.system( "strip " + ncbipyFile )
            os.system( "cp " + ncbipyFile + " " + targetDir )

            # Compile the python code
            os.system( "touch " + targetDir + "__init__.py" )
            os.system( "cd " + targetDir + "; ./ncbipy -c 'import ncbi'" )
            os.system( "cd " + targetDir + "; ./ncbipy -c 'import __init__'" )

    return 0


def generateVersionFile( version ):
    """ generates a file with a version constant """

    try:
        f = open( 'version.i', 'w' )
        f.write( '/*\n' )
        f.write( ' * Auto generated file\n' )
        f.write( ' */\n' )
        f.write( '\n' )
        f.write( '%constant version = "' + version + '";\n' )
        f.close()
    except:
        print >> sys.stderr, "Error generating version.i with a version " \
                             "constant. Exiting"
        return 2

    return 0


def buildWrapperLibrary( buildDir, verbose ):
    """
    Replaces the fake 0-length source code and
    runs make in the proper directory
    Should return non zero in case of errors
    """

    # For all target languages
    for lang in getTargetLanguages():
        srcFileName = lang + "_wrappers.cpp"
        targetCppFile = lang + "_wrappers/" + srcFileName
        retCode = os.system( "rm -rf " + targetCppFile )
        retCode += os.system( "cp " + srcFileName + " " + targetCppFile )
        if retCode != 0:
            print >> sys.stderr, "Error replacing a dummy source code file " \
                                 "with the generated one for " + lang +      \
                                 ". Source file: " + srcFileName +           \
                                 " Target file: " + targetCppFile + "\nExiting."
            return 2

        # Run make in the proper dir
        wrapperBuildDir = buildDir + "internal/cppcore/toolkit_wrappers/" + \
                          lang + "_wrappers"
        cmdLine = "cd " + wrapperBuildDir + "; make all_p"
        if verbose:
            print "Wrapper building command line: " + cmdLine

        retCode = os.system( cmdLine )
        if retCode != 0:
            print >> sys.stderr, "Error making wrapper for " + lang + \
                                 "\n Exiting."
            return 2

    return 0


def runSWIG( incDir, buildDir, verbose ):
    """
    Runs SWIG to generate the cpp wrapper file
    Should return non zero in case of errors
    """

    # Pick up the -D keys which were used to build the toolkit libraries
    makefileInPath = buildDir + "Makefile.mk"
    if not os.path.exists( makefileInPath ):
        print >> sys.stderr, "Cannot find the build Makefile.mk. " \
                             "Expected it here: " + makefileInPath + ". Exiting"
        return 2

    defines = []
    cppFlags = safeRun( [ "grep", "CONF_CPPFLAGS", makefileInPath ] ).strip()
    parts = cppFlags.split()
    for part in parts:
        part = part.strip()
        if part.startswith( "-D" ):
            defines.append( part )

    # Form the -I directives
    includes = []
    includes.append( "-I" + incDir[ :-1 ] )
    includes.append( "-I" + buildDir.replace( "/build/", "/inc" ) )
    # includes.append( "-I/usr/include" )

    # For all target languages
    for lang in getTargetLanguages():
        if verbose:
            print "Running SWIG for " + lang
        cmdLine  = "swig -w312 -v -ignoremissing -c++ -importall -" + lang
        cmdLine += " " + " ".join( defines )
        cmdLine += " " + " ".join( includes )
        cmdLine += " -o " + lang + "_wrappers.cpp"
        cmdLine += " ncbi.i"
        if verbose:
            print "SWIG command line: " + cmdLine

        retCode = os.system( cmdLine )
        if retCode != 0:
            print >> sys.stderr, "Error running SWIG for " + lang + ". Exiting"
            return 2
        print "SWIG run completed for " + lang
        print "-------------------------------"

    return 0



def getTargetLanguages():
    """ Provides a list of the target languages """

    # default language is Python
    return os.environ.get( "NCBI_SWIG_LANG",
                           "python" ).strip().lower().split()



def checkCoverage( acceptedHeaders ):
    """
    Checks that all the toolkit headers are covered
    Should return non zero in case of errors
    """

    print "COVERAGE CHECK HAS NOT BEEN IMPLEMENTED YET"
    return 0


def excludeHeaders( headers, cppDir, verbose ):
    """
    opens a file holds a list of exceptions and exclude some toolkit headers
    also strips a leading part to /c++ inclusive
    """

    patterns = []
    try:
        f = open( "ignore_toolkit_headers", "r" )
        for line in f:
            line = line.strip()
            if line == "" or line.startswith( "#" ):
                continue
            patterns.append( re.compile( line ) )
            if verbose:
                print "Exclude filter found: " + line
        f.close()
    except:
        print "Warning: error processing " \
              "'ignore_toolkit_headers' file. Ignoring..."

    acceptedHeaders = []
    for item in headers:
        header = item.replace( cppDir + "include/", "" )
        matched = False
        for pattern in patterns:
            if pattern.match( header ):
                matched = True
                break
        if not matched:
            acceptedHeaders.append( header )
        else:
            if verbose:
                print "Toolkit header to be excluded: " + header

    return acceptedHeaders


def toolkitHeaders( path, headers ):
    """ Builds a list of all the toolkit header files """

    for item in os.listdir( path ):
        dirItem = path + "/" + item
        if os.path.isdir( dirItem ):
            toolkitHeaders( dirItem, headers )
            continue
        if os.path.isfile( dirItem ):
            if dirItem.endswith( ".hpp" ) or dirItem.endswith( ".h" ):
                headers.append( dirItem )
    return


def parserError( parser, message ):
    """ Prints the message and help on stderr """

    sys.stdout = sys.stderr
    print message
    parser.print_help()
    return 2


def safeRun( commandArgs ):
    """
    Runs the given command and reads the output
    """

    errTmp = tempfile.mkstemp()
    errStream = os.fdopen( errTmp[0] )
    process = Popen( commandArgs, stdin = PIPE,
                     stdout = PIPE, stderr = errStream )
    process.stdin.close()
    processStdout = process.stdout.read()
    process.stdout.close()
    errStream.seek( 0 )
    err = errStream.read()
    errStream.close()
    os.unlink( errTmp[1] )
    process.wait()

    # 'grep' return codes:
    # 0 - OK, lines found
    # 1 - OK, no lines found
    # 2 - Error occured
    if process.returncode != 0 and commandArgs[0] != "grep":
        raise Exception( "Error in '%s' invocation: %s" % \
                         (commandArgs[0], err) )
    return processStdout


def getExceptionInfo():
    """
    The function formats the exception and returns the string which
    could be then printed or logged
    """

    excType, value, tback = sys.exc_info()
    msg = str( value )
    if len( msg ) == 0:
        msg = "There is no message associated with the exception."
    if msg.startswith( '(' ) and msg.endswith( ')' ):
        msg = msg[1:-1]

    try:
        tbInfo = format_tb( tback )
        tracebackInfoMsg = "Traceback information:\n" + "".join( tbInfo )
    except:
        tracebackInfoMsg = "No traceback information available"

    return "Exception is caught. " + msg + "\n" + tracebackInfoMsg



# The script execution entry point
if __name__ == "__main__":
    returnCode = 2
    try:
        returnCode = main()
    except:
        print >> sys.stderr, getExceptionInfo()

    if returnCode != 0:
        returnCode = 2
    sys.exit( returnCode )

